import numpy as np
import lmfit as lm


def fitPeak(data,xleft,xright,peakType='Gaussian',constant=False):
    '''
    Fit data to a peak model with the provided bounds and the addition of a constant background term
    
    Parameters
    --------
    data : array_like
        Input 2D array containing the data to fit.
    xleft : float
        Left boundary of the data to fit.
    xright : float
        Right boundary of the data to fit.
    peakType : str, optional
        The type of peak to use for the fit. The default choice is a Gaussian model.
        All models are listed below.
    constant : bool, optional
        Whether to add a constant background value to the fit.


    Returns
    --------
    fitResult : lmfit ModelResult object
        The ModelResult contains the fitting results created by lmfit. 
    '''
    
    if peakType not in lm.models.lmfit_models.keys():
        print('Peak type not available. Options are:')
        for keys in lm.models.lmfit_models.keys():
            print(keys)
        return -1
    
    datax = data[:,0]
    datay = data[:,1]
    
    modPeak = lm.models.lmfit_models[peakType]()
    modConst = lm.models.ConstantModel()
    if constant:
        modCombined = modPeak + modConst
    else:
        modCombined = modPeak
    
    dataLocations = np.where((datax >= xleft) & (datax <= xright))[0]
    
    fitx = datax[dataLocations]
    fity = datay[dataLocations]
    
    
    params = modConst.guess(fity,x=fitx)
    params+= modPeak.guess(fity,x=fitx)
    
    params['c'].set(min=0,max=np.max(fity))
    
    fitResult = modCombined.fit(fity,params,x=fitx)
    
    return fitResult


def find_max(fitfunction,bounds):
    '''
    Finds the maximum value of a Gaussian shaped function determined through a fit

    Parameters
    ---------
    fitfunction : lmfit fitted class
        Fitting result from lmfit
    
    bounds : array_like
        A list of the bounds where the peak is located.

    Return
    --------
    out : array with xvalue, maxvalue, maxvalueunc


    Example
    --------
    modGaus = lmfit.models.Gaussian()
    result = modGaus.fit(y,params,x=x)
    maximum = find_max(result,[1,2])
        '''
    from scipy.optimize import minimize_scalar
    
    def function(xvalue):
        return -fitfunction.eval(x=xvalue)
    xvalue = minimize_scalar(function,bounds=bounds,method='bounded').x
    maxvalue = fitfunction.eval(x=xvalue)
    maxvalueunc = fitfunction.eval_uncertainty(x=xvalue)[0]
    out = np.array([xvalue, maxvalue, maxvalueunc])
    return out


def convertToUncFloat(paramResult):
    '''
    Converts a parameter output from lmfit to a uncertainties ufloat.

    Parameters
    ------
    paramResult : Paramtere object from lmfit
            Input the parameter after a fit to retrive the value
            and uncertainty into a ufloat.

    Returns
    ------
    out : ufloat
        Returns a uncertainties ufloat value.
    '''
    from uncertainties import ufloat

    out = ufloat(paramResult.value,paramResult.stderr)

    return out


# from numpy import array, sqrt, asarray, argmin, abs, arange, zeros, sum, sinc, pi, max, column_stack
import numpy as np
from scipy.integrate import simpson
import uncertainties as unc
# import os
# from pathlib import Path
# from varname import nameof


gr = (1 + np.sqrt(5))/2 #Defines the golden ratio



def find_nearest(array, value):
    '''
    Finds the nearest value in an array and its position

    Input
    -----------
    array, value


    Return 
    ----------
    Nearest value
    '''
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx, array[idx]



def discrete_integral(data):
    '''
    Calculate a continuous discrete integral of data using the simpson method

    Parameters
    ----------
    data : array_like
        The input array to be integrated.

    Returns
    -------
    dataInt : float value of integration

    '''
    dataInt = data * 0 
    for i in range(1,len(data)+1):
        dataInt[i-1,:] = np.array([data[i-1,0],simpson(data[:i,1],data[:i,0])])
    return dataInt



def fourierbesseltransform(q,int1,unpack=None):
    '''
    Discrete fourier bessel transform for conversion between S(Q) and g(r)

    Parameters
    --------
    q : array_like
        Input array of Q (or r) values
    int1 : array_like
        Input array of the structure factor S values (or PDF values)
    unpack : bool, optional
        If unpack is True, the result will be output as a tuple to 
        more easily define separate variables from the result

    Returns 
    --------
    r : ndarray
        array of the real space distance (or Q) values.
    ft : ndarray
        fourier transform result. Returns the PDF (or S) values.
    '''    
    #First define delta Q
    dq = q[1]-q[0] #Q or r is assumed to be equidistant
    
    q2 = q**2
    
    intq2 = int1 * q2 #Multiply by Q^2 before
    
    nq = len(q)
    
    r = np.arange(nq) * np.pi / np.max(q) #Define the values of r
    nr = len(r)
    ft = np.zeros(nr) #Create empty array (detault is float64)
    
    for i in range(nr):
        x = q * r[i]
        ft[i] = sum(intq2 * np.sinc(x/np.pi)) #np.sinc = sin(pi*x)/(pi*x)
    
    ft *= dq
    
    if unpack is None:
        return np.column_stack((r, ft))
    else:
        return r, ft
    
    
    
# def SaveToFile(fname,*data,header=None,directory=None,**kwargs):
    
#     #Function to save numpy arrays to a file
#     currDir = Path(os.getcwd())
#     if directory is None:
#         outputDir = currDir
#     else:
#         outputDir = Path(directory)
#         if not outputDir.exists():
#             response = input(f'The directory:\n{outputDir}\n does not exist. Should it be created? y/n')
#             if response == 'y':
#                 outputDir.mkdir(parents=True, exist_ok=True)
#                 print('Directory created.')
#             else:
#                 print('Function aborted.')
#                 return
#     if '.' not in fname:
#         fname += '.dat'
#     outputFile = outputDir / str(fname)

#     openFile = open(outputFile,'w')
#     if header is not None:
#         if '#' not in header[0]:
#             header = ['#' + s for s in header]
#         openFile.writelines(header)
    
#     np.savetxt(openFile,np.column_stack(data),**kwargs)
#     openFile.close()
#     print(f'The data has been successfully save as: \n {outputFile}')
#     return


def moving_average(x, w):
    '''
    Computes the moving (or running or rolling) average of input data.

    Parameters
    ---------
    x : array_like
        Input data to perform moving average on
    w : int
        The window (or box) size to be averaged

    Returns
    -------
    out : ndarray
        The resulting array after performing the moving average. 
        The array will have a length of n-w+1
    '''
    out = np.convolve(x, np.ones(w), 'valid') / w
    return out



def convertToUncFloat(paramResult):
    '''
    Returns a ufloat from the inputed lmfit parameter class

    Parameters 
    --------
    paramResult : lmfit parameter class
                  Fitting parameters after using lmfit to perform fits.

    Returns
    --------
    out : uncertainties float (ufloat)
          The float value containing the uncertainty calculated from the fitting procedure.
    '''
    out = unc.ufloat(paramResult.value,paramResult.stderr)
    return out



def bin_data(dataArray_x,dataArray_y,minValue,maxValue,dataPoints,unpack=False,method='mean',density=False):
    '''
    Returns rebinned x and y arrays

    Parameters
    ---------
    dataArray_x : array_like
                  The x-values of the array to be rebinned
    dataArray_y : array_like
                  The y-values of the array to be rebinned
    minValue : value
               The lower bound x-value of the new rebinned data
    maxValue : value
               The upper bound x-value of the new rebinned data
    dataPoint : value
                The number of bins to combine the data into
    unpack : bool, optional
             If unpack is True, the result will be output as a tuple to 
             more easily define separate variables from the result
    method : str, optional
             Choose either "sum" or "mean" for the method of combining data
    density : bool, optional
              Choose True to divide the data by the bin size creating a density = occurance/bin width

    Returns
    --------
    binnedArray_x : ndarray
                    New x-value bins
    binnedArray_y : ndarray
                    New y-values after binning

    Example
    ---------
    x = np.array([0,1,2,3,4,5,6,7,8,9,10])
    y = np.array([5,6,3,4,6,8,9,6,3,4,5])

    output = bin_data(x,y,0,10,5)
    print(output)

    [[1.  4.5]
    [3.  5. ]
    [5.  8.5]
    [7.  4.5]
    [9.  4.5]]
    
    
    '''
    from uncertainties.unumpy import uarray

    if method == 'mean':
        func = np.mean
    elif method == 'sum':
        func = np.sum
    else:
        print('Method must be either "mean" or "sum".')
        return -1

    binWidths = np.linspace(minValue,maxValue,dataPoints+1)
    binnedArray_x = np.zeros(dataPoints)
    
    if dataArray_y.dtype == 'O':
        binnedArray_y = uarray(range(dataPoints),range(dataPoints)) * 0.0
    else:
        binnedArray_y = np.zeros(dataPoints)
    
    for index in range(dataPoints):
        left = binWidths[index]
        right = binWidths[index+1]
        binnedArray_x[index] = np.mean([left,right])

        if density: 
            norm = (right - left)
        else:
            norm = 1
        
        locations = np.where((dataArray_x > left) & (dataArray_x <= right))[0]
        if len(locations) != 0:
            binnedArray_y[index] = func(dataArray_y[locations]) / norm
        else:
            if dataArray_y.dtype == 'O':
                binnedArray_y[index] = unc.ufloat(0.0,0.0)
            else:
                binnedArray_y[index] = 0.0
        
        
    if unpack:
        return binnedArray_x, binnedArray_y
    else:
        return np.column_stack((binnedArray_x, binnedArray_y))
    


    
def bin_ndarray(ndarray, new_shape, operation='sum'):
    """
    Bins an ndarray in all axes based on the target shape, by summing or
        averaging.

    Number of output dimensions must match number of input dimensions and 
        new axes must divide old ones.

    Example
    -------
    >>> m = np.arange(0,100,1).reshape((10,10))
    >>> n = bin_ndarray(m, new_shape=(5,5), operation='sum')
    >>> print(n)

    [[ 22  30  38  46  54]
     [102 110 118 126 134]
     [182 190 198 206 214]
     [262 270 278 286 294]
     [342 350 358 366 374]]

    """
    operation = operation.lower()
    if not operation in ['sum', 'mean']:
        raise ValueError("Operation not supported.")
    if ndarray.ndim != len(new_shape):
        raise ValueError("Shape mismatch: {} -> {}".format(ndarray.shape,
                                                           new_shape))
    compression_pairs = [(d, c//d) for d,c in zip(new_shape,
                                                  ndarray.shape)]
    flattened = [l for p in compression_pairs for l in p]
    ndarray = ndarray.reshape(flattened)
    for i in range(len(new_shape)):
        op = getattr(ndarray, operation)
        ndarray = op(-1*(i+1))
    return ndarray


import pandas as pd
import numpy as np
# from importlib import resources
# import io


def atomic_form_factor_constants():
    '''
    Outputs a DataFrame of the coefficients for the analytical approximation to the atomic form factors. 
    The coefficients were taken from the International Tables for Crystallography at:
    http://it.iucr.org/Cb/ch6o1v0001/

    Parameters
    -------
    nothing


    Returns 
    --------
    aff_DF : Pandas DataFrame
        contains the coefficients to use the analytical approximation of the atomic form factors. 
    '''
    from pkg_resources import resource_stream

    stream = resource_stream(__name__,'Data/AtomicFormFactorConstants.csv')
    aff_DF = pd.read_csv(stream)
    return aff_DF



def atomic_form_factor(atom,QList,inputCoeff=None):
    '''
    Returns the Q dependent atomic form factor for each of the elements and ions. 
    The coefficients were taken from the International Tables for Crystallography at:
    http://it.iucr.org/Cb/ch6o1v0001/

    Parameters
    -------
    atom : str
        Put the name of the atom (or ion) for the atomic form factor.
        Ions have their charge put directly after the atom in the form
        X#+/-
        For example: Pt4+ or H1-
    
    QList : array_like
        The array of Q values that the atomic form factor will be calculated for.

    inputCoeff : array_like, optional
        Optionally input the coefficients manually rather than use the values from the table.
    '''

    aff_DF = atomic_form_factor_constants()

    if atom not in aff_DF['element'].values:
        print(f'The atom \'{atom}\' is not available.\nThe available atoms are:\n','\t'.join(map(str,aff_DF['element'].values)))
    else:
        if inputCoeff is None:
            coeff_values = aff_DF[aff_DF['element'] == atom].iloc[:,2:-2].values[0]
        else:
            coeff_values = inputCoeff
        sumData = 0.0
        for i in range(0,len(coeff_values)-1,2):
            sumData += coeff_values[i] * np.exp(-coeff_values[i+1] * (QList/(4*np.pi))**2)
        sumData += coeff_values[-1]
        return sumData


def neutron_scattering_lengths(rawTable=False):
    '''
    Returns a DataFrame of all the neutron scattering lengths. 
    The neutron scattering lengths were taken from:
    https://www.nist.gov/ncnr/neutron-scattering-lengths-list

    Parameters
    --------
    rawTable : bool, options
        Set true to output the raw table imported from the above link.
        Otherwise it will import a table with symbols removed (errors +/- values etc.)
        and corrected data types.

    Returns
    -------
    nsl_DF : DataFrame
        Contains the neutron scattering lengths of all elements and isotopes.

    Column    Unit    Quantity
    1         ---     Isotope
    2         ---     Natural abundance (for radioisotopes the half-life is given instead)
    3         fm      bound coherent scattering length
    4         fm      bound incoherent scattering length
    5         barn    bound coherent scattering cross section
    6         barn    bound incoherent scattering cross section
    7         barn    total bound scattering cross section
    8         barn    absorption cross section for 2200 m/s neutrons

    Note: 1 fm = 1E-15 m, 1 barn = 1E-24 cm^2, scattering legnths and cross sections in parenthesis are uncertainties
    '''

    from pkg_resources import resource_stream

    if rawTable:
        stream = resource_stream(__name__,'Data/NeutronScatteringLengths.csv')
        nsl_DF = pd.read_csv(stream)
        return nsl_DF


    stream = resource_stream(__name__,'Data/NeutronScatteringLengths_Corrected.csv')
    nsl_DF = pd.read_csv(stream).astype({'Isotope':'string',
                                        'Conc':np.float64,
                                        'Coh b':np.complex64,
                                        'Inc b':np.complex64,
                                        'Coh xs':np.float64,
                                        'Inc xs':np.float64,
                                        'Scatt xs':np.float64,
                                        'Abs xs':np.float64})
    return nsl_DF